/* Name: infrahid.c
 * Project: InfraHID; CIR to HID mouse/keyboard converter
 * Author: Alex Badea
 * Creation Date: 2007-01-14
 * Tabsize: 8
 * Copyright: (c) 2007 Alex Badea <vamposdecampos@gmail.com>
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"
#include "irmp.h"


#define nop()	__asm__ __volatile__ ("nop");
#define reset()	__asm__ __volatile__ ("rjmp 0");

#define led_on()	do { PORTB |= _BV(0); } while (0)
#define led_off()	do { PORTB &= ~_BV(0); } while (0)
#define led_toggle()	do { PORTB ^= _BV(0); } while (0)

#define ABS(x)	((x) < 0 ? -(x) : (x))

/* ------------------------------------------------------------------------- */

static const char hidReportDescriptor0[] PROGMEM = {
	/* partial keyboard */
	0x05, 0x01,	/* Usage Page (Generic Desktop), */
	0x09, 0x06,	/* Usage (Keyboard), */
	0xA1, 0x01,	/* Collection (Application), */
	0x85, 0x01,		/* Report Id (1) */
	0x95, 0x04,		/* Report Count (4), */
	0x75, 0x08,		/* Report Size (8), */
	0x15, 0x00,		/* Logical Minimum (0), */
	0x25, 0x65,		/* Logical Maximum(101), */
	0x05, 0x07,		/* Usage Page (Key Codes), */
	0x19, 0x00,		/* Usage Minimum (0), */
	0x29, 0x65,		/* Usage Maximum (101), */
	0x81, 0x00,		/* Input (Data, Array),               ;Key arrays (4 bytes) */
	0xC0,		/* End Collection */

	/* mouse */
	0x05, 0x01,	/* Usage Page (Generic Desktop), */
	0x09, 0x02,	/* Usage (Mouse), */
	0xA1, 0x01,	/* Collection (Application), */
	0x09, 0x01,	/*   Usage (Pointer), */
	0xA1, 0x00,	/*   Collection (Physical), */
	0x05, 0x09,		/* Usage Page (Buttons), */
	0x19, 0x01,		/* Usage Minimum (01), */
	0x29, 0x03,		/* Usage Maximun (03), */
	0x15, 0x00,		/* Logical Minimum (0), */
	0x25, 0x01,		/* Logical Maximum (1), */
	0x85, 0x02,		/* Report Id (2) */
	0x95, 0x03,		/* Report Count (3), */
	0x75, 0x01,		/* Report Size (1), */
	0x81, 0x02,		/* Input (Data, Variable, Absolute), ;3 button bits */
	0x95, 0x01,		/* Report Count (1), */
	0x75, 0x05,		/* Report Size (5), */
	0x81, 0x01,		/* Input (Constant),                 ;5 bit padding */
	0x05, 0x01,		/* Usage Page (Generic Desktop), */
	0x09, 0x30,		/* Usage (X), */
	0x09, 0x31,		/* Usage (Y), */
	0x15, 0x81,		/* Logical Minimum (-127), */
	0x25, 0x7F,		/* Logical Maximum (127), */
	0x75, 0x08,		/* Report Size (8), */
	0x95, 0x02,		/* Report Count (2), */
	0x81, 0x06,		/* Input (Data, Variable, Relative), ;2 position bytes (X & Y) */
	0xC0,		/*   End Collection, */
	0xC0,		/* End Collection */
};


static const char configDescriptor[] PROGMEM = {	/* USB configuration descriptor */
	9,			/* sizeof(usbDescriptorConfiguration): length of descriptor in bytes */
	USBDESCR_CONFIG,	/* descriptor type */
	9 + 1 * (9 + 9 + 7), 0,
	/* total length of data returned (including inlined descriptors) */
	1,			/* number of interfaces in this configuration */
	1,			/* index of this configuration */
	0,			/* configuration name string index */
	USBATTR_BUSPOWER,	/* attributes */
	USB_CFG_MAX_BUS_POWER / 2,	/* max USB current in 2mA units */

/* --- interface 0 --- */
	9,			/* sizeof(usbDescrInterface): length of descriptor in bytes */
	USBDESCR_INTERFACE,	/* descriptor type */
	0,			/* index of this interface */
	0,			/* alternate setting for this interface */
	1,			/* endpoints excl 0: number of endpoint descriptors to follow */
	3,			/* 3=HID class */
	0,			/* subclass: 0=none, 1=boot */
	0,			/* protocol: 0=none, 1=keyboard, 2=mouse */
	0,			/* string index for interface */

	/* HID descriptor */
	9,			/* sizeof(usbDescrHID): length of descriptor in bytes */
	USBDESCR_HID,		/* descriptor type: HID */
	0x01, 0x01,		/* BCD representation of HID version */
	0x00,			/* target country code */
	0x01,			/* number of HID Report (or other HID class) Descriptor infos to follow */
	0x22,			/* descriptor type: report */
	sizeof(hidReportDescriptor0), 0,	/* total length of report descriptor */

	/* endpoint descriptor for endpoint 1 */
	7,			/* sizeof(usbDescrEndpoint) */
	USBDESCR_ENDPOINT,	/* descriptor type = endpoint */
	0x81,			/* IN endpoint number 1 */
	0x03,			/* attrib: Interrupt endpoint */
	8, 0,			/* maximum packet size */
	USB_CFG_INTR_POLL_INTERVAL,	/* in ms */
};

uchar usbFunctionDescriptor(usbRequest_t * rq)
{
	uchar *p = 0, len = 0;

	DBG1(0xa8, &rq->wValue.bytes[1], 1);

#if 0	/* default descriptor from driver */
	if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
		p = (uchar *) deviceDescriptor;
		len = sizeof(deviceDescriptor);
	} else
#endif
	if (rq->wValue.bytes[1] == USBDESCR_CONFIG) {
		p = (uchar *) configDescriptor;
		len = sizeof(configDescriptor);
	} else if (rq->wValue.bytes[1] == USBDESCR_HID) {
		p = (uchar *) (configDescriptor + 18);
		len = 9;
	} else if (rq->wValue.bytes[1] == USBDESCR_HID_REPORT) {
		p = (uchar *) hidReportDescriptor0;
		len = sizeof(hidReportDescriptor0);
	}
	usbMsgPtr = p;
	return len;
}


/* ------------------------------------------------------------------------- */

static uchar inputBuffer1[5] = { 1, 0,0,0,0, };		/* kbd */
static uchar inputBuffer2[4] = { 2, 0,0,0, };		/* mouse */
static uchar idleRate = 0;				/* in 4 ms units */
static uchar tx_kbd = 0;
static uchar tx_mouse = 0;

uchar usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = (void *) data;

	DBG1(0xa0, data, 8);

	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {	/* class request type */
		if (rq->bRequest == USBRQ_HID_GET_REPORT) {
			/* wValue: ReportType (highbyte), ReportID (lowbyte) */
			if (rq->wValue.bytes[0] == 1) {
				usbMsgPtr = inputBuffer1;
				return sizeof(inputBuffer1);
			} else if (rq->wValue.bytes[0] == 2) {
				usbMsgPtr = inputBuffer2;
				return sizeof(inputBuffer2);
			}
			return 0;
		} else if (rq->bRequest == USBRQ_HID_SET_REPORT) {
			/* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* we have no output/feature reports */
			return 0;
		} else if (rq->bRequest == USBRQ_HID_GET_IDLE) {
			usbMsgPtr = &idleRate;
			return 1;
		} else if (rq->bRequest == USBRQ_HID_SET_IDLE) {
			idleRate = rq->wValue.bytes[1];
		}
		return 0;
	}

	return 0;
}

/* ------------------------------------------------------------------------- */


/* ir_state values */
#define IR_IDLE		0
#define IR_HEADER	1
#define IR_DATA_HI	2
#define IR_DATA_LO	3
#define IR_DISCARD	4

/* mouse acceleration values */
#define IR_ACCEL_DEF	2
#define IR_ACCEL_INC_LO	1
#define IR_ACCEL_INC_HI	2
#define IR_ACCEL_MAX	50

/* remote codes */
#define IRCODE_UP	0x30cf
#define IRCODE_DOWN	0x08f7
#define IRCODE_LEFT	0x10ef
#define IRCODE_RIGHT	0x20df
#define IRCODE_FULL	0xc03f
#define IRCODE_FINE_LT	0x9867
#define IRCODE_FINE_RT	0x18e7
		    
#define IR_TIMEOUT	6	/* in 21.8ms intervals */

struct code_entry {
	unsigned char remote_protocol;
	unsigned char remote_command;
	uchar hid_code;
} code_map[] = {
	{ 0x02, 0x54, 30 },	/* 1 */
	{ 0x02, 0x55, 31 },	/* 2 */
	{ 0x02, 0x56, 32 },	/* 3 */
	{ 0x02, 0x57, 33 },	/* 4 */
	{ 0x02, 0x58, 34 },	/* 5 */
	{ 0x02, 0x59, 35 },	/* 6 */
	{ 0x02, 0x5A, 36 },	/* 7 */
	{ 0x02, 0x5B, 37 },	/* 8 */
	{ 0x02, 0x5C, 38 },	/* 9 */
	{ 0x02, 0x5D, 39 },	/* 0 */

	{ 0x20, 0x00, 0xf0 },	/* Up */
	{ 0x20, 0x01, 0xf1 },	/* Down */
	{ 0x20, 0x02 ,0xf2 },	/* Left */
	{ 0x20, 0x03 ,0xf3 },	/* Right */
	{ 0x20, 0x04, 0xf4 },	/* Up */
	{ 0x20, 0x05, 0xf5 },	/* Down */
	{ 0x20, 0x06 ,0xf6 },	/* Left */
	{ 0x20, 0x07 ,0xf7 },	/* Right */
	{ 0x20, 0x08, 0xf8 },	/* Up */
	{ 0x20, 0x09, 0xf9 },	/* Down */
	{ 0x20, 0x0A ,0xfa },	/* Left */
	{ 0x20, 0x0B ,0xfb },	/* Right */
	{ 0x20, 0x0C, 0xfc },	/* Up */
	{ 0x20, 0x0D, 0xfd },	/* Down */
	{ 0x21, 0x4E ,0xfe },	/* Left */
	{ 0x22, 0x2F ,0xf4 },	/* Right */
};



//static const uchar  code_map[NUM_CODES][3] PROGMEM = {
//{ 0x02, 0x54, 30 },	/* 1 */
//{ 0x02, 0x55, 31 },	/* 2 */
//{ 0x02, 0x56, 32 },	/* 3 */
//{ 0x02, 0x57, 33 },	/* 4 */
//{ 0x02, 0x58, 34 },	/* 5 */
//{ 0x02, 0x59, 35 },	/* 6 */
//{ 0x02, 0x5A, 36 },	/* 7 */
//{ 0x02, 0x5B, 37 },	/* 8 */
//{ 0x02, 0x5C, 38 },	/* 9 */
//{ 0x02, 0x5D, 39 },	/* 0 */
//
//{ 0x20, 0x00, 0xf0 },	/* Up */
//{ 0x20, 0x01, 0xf1 },	/* Down */
//{ 0x20, 0x02 ,0xf2 },	/* Left */
//{ 0x20, 0x03 ,0xf3 },	/* Right */
//{ 0x20, 0x04, 0xf4 },	/* Up */
//{ 0x20, 0x05, 0xf5 },	/* Down */
//{ 0x20, 0x06 ,0xf6 },	/* Left */
//{ 0x20, 0x07 ,0xf7 },	/* Right */
//{ 0x20, 0x08, 0xf8 },	/* Up */
//{ 0x20, 0x09, 0xf9 },	/* Down */
//{ 0x20, 0x0A ,0xfa },	/* Left */
//{ 0x20, 0x0B ,0xfb },	/* Right */
//{ 0x20, 0x0C, 0xfc },	/* Up */
//{ 0x20, 0x0D, 0xfd },	/* Down */
//{ 0x21, 0x4E ,0xfe },	/* Left */
//{ 0x22, 0x2F ,0xf4 },	/* Right */
//};
#define NUM_CODES	(sizeof(code_map) / sizeof(struct code_entry))

//#define NUM_CODES	(sizeof(code_map) / sizeof(code_entry))


static unsigned char ir_timeout = 0;

static void hid_clear(void)
{
	inputBuffer2[1] = 0;	/* mouse buttons */
	inputBuffer2[2] = 0;	/* X axis delta */
	inputBuffer2[3] = 0;	/* Y axis delta */

	inputBuffer1[1] = 0;	/* key */
}

void ir_input(IRMP_DATA * irmp_data)
{
	
	char repeat = 0;
	uchar k;

//	if (ir_data == 0xa857)	/* "scan" button */
//		reset();	/* XXX HACK */

	if(ir_timeout > 0){
		repeat = 1;
	}else{
		repeat=0;
	}

	if (!usbInterruptIsReady()) {
		//led_on();
		return;		/* still sending interrupt packet */
	}
	
	hid_clear();
	
/*	if(irmp_data->address==0 && irmp_data->command==0 && irmp_data->flags==0 && irmp_data->protocol==0){
		tx_mouse = 1;
		tx_kbd = 1;
		return;
	}*/


 	/*if (repeat) {
 		if (ir_accel < IR_ACCEL_MAX / 2)
 			ir_accel += IR_ACCEL_INC_LO;
 		else if (ir_accel < IR_ACCEL_MAX)
 			ir_accel += IR_ACCEL_INC_HI;
 	} else {
 		ir_accel = IR_ACCEL_DEF;
 	}*/


	for (k = 0; k < NUM_CODES; k++) {
		uchar action;
		
		if ((irmp_data->command == code_map[k].remote_command) && (irmp_data->protocol == code_map[k].remote_protocol))
		{ 		
			action = code_map[k].hid_code;
			
			
			//if (action == 0xfa)
			//inputBuffer2[3] = -3; /* up */
			//else if (action == 0xfb)
			//inputBuffer2[3] = 3; /* down */
			//else if (action == 0xfc)
			//inputBuffer2[2] = -3; /* left */
			//else if (action == 0xfd)
			//inputBuffer2[2] = 3; /* right */
			

			if (action == 0xf0){       /* oben */
				inputBuffer2[3] = -4;
			}else if (action == 0xf1){
				inputBuffer2[2] = 1; 
				inputBuffer2[3] = -3; 
			}else if (action == 0xf2){
				inputBuffer2[2] = 2; 
				inputBuffer2[3] = -2; 
			}else if (action == 0xf3){
				inputBuffer2[2] = 3; 
				inputBuffer2[3] = -1; 
			}else if (action == 0xf4){ /* rechts */
				inputBuffer2[2] = 4;
			}else if (action == 0xf5){
				inputBuffer2[2] = 3;
				inputBuffer2[3] = 1;
			}else if (action == 0xf6){
				inputBuffer2[2] = 2;
				inputBuffer2[3] = 2;	
			}else if (action == 0xf7){
				inputBuffer2[2] = 1;
				inputBuffer2[3] = 3;
			}else if (action == 0xf8){ /* unten */
				inputBuffer2[3] = 4;
			}else if (action == 0xf9){
				inputBuffer2[2] = -1;
				inputBuffer2[3] = 3;
			}else if (action == 0xfa){
				inputBuffer2[2] = -2;
				inputBuffer2[3] = 2;
			}else if (action == 0xfb){
				inputBuffer2[2] = -3;
				inputBuffer2[3] = 1;
			}else if (action == 0xfc){ /* links */
				inputBuffer2[2] = -4;
			}else if (action == 0xfd){
				inputBuffer2[2] = -3;
				inputBuffer2[3] = -1;
			}else if (action == 0xfe){
				inputBuffer2[2] = -2;
				inputBuffer2[3] = -2;
			}else if (action == 0xff){
				inputBuffer2[2] = -1;
				inputBuffer2[3] = -3;
			}						
			//else if ((action & 0xf8) == 0xf0)
			//	inputBuffer2[1] |= 1 << (action & 7); /* click */
			else {
				inputBuffer1[1] = action;
				tx_kbd = 1;
				usbSetInterrupt(inputBuffer1, sizeof(inputBuffer1));	/* send kbd event */
				break;
			}
			tx_mouse = 1;
			usbSetInterrupt(inputBuffer2, sizeof(inputBuffer2));	/* send mouse event */
			break;
		}	
	}

	ir_timeout = IR_TIMEOUT;
	//led_toggle();
}


/* this may well be SIGNAL(SIG_OVERFLOW0), but we want it called at well-defined times */
void idle_timer(void)
{	
	
	static uchar idleCounter = 0;

	/* idle reports, if requested */
	if (idleRate) {
		if (idleCounter > 4) {
			idleCounter -= 5;
		} else {
			idleCounter = idleRate;
			tx_kbd = 1;
			tx_mouse = 1;
		}
	}
	
	
			
	/* process button releases */
/*	if(ir_timeout>IR_TIMEOUT){
		ir_timeout=IR_TIMEOUT;
	}
	if(ir_timeout<0){
		ir_timeout=0;
	}	*/	
			
	if (ir_timeout>0) {
		--ir_timeout;
		if (ir_timeout==0) {
			hid_clear();
			tx_kbd = 1;
			tx_mouse = 1;
			//led_on();
			led_toggle();	
		}
	}
}

void send_packets(void)
{
	/* send pending packets */
	if (usbInterruptIsReady()) {
		if (tx_kbd) {
			usbSetInterrupt(inputBuffer1, sizeof(inputBuffer1));	/* send kbd event */
			tx_kbd = 0;
		} else if (tx_mouse) {
			usbSetInterrupt(inputBuffer2, sizeof(inputBuffer2));	/* send mouse event */
			tx_mouse = 0;
		}
	}
}

/* ------------------------------------------------------------------------- */

/* "more thorough" reset: loop 4 times outputting SE0 then idle.
 * This works around some intermitent connection problems. */
void usbReset(void)
{
	uchar i, j, k;

	for (k = 0; k < 8; k++) {
		/* on even iterations, output SE0 */
		if (!(k & 1)) {
			PORTD &= ~USBMASK;	/* no pullups on USB pins */
			DDRD |= USBMASK;	/* output SE0 for USB reset */
		} else {
			DDRD &= ~USBMASK;	/* set USB data as inputs */
		}

		/* delay */
		j = 0;
		while (--j) {		/* USB Reset by device only required on Watchdog Reset */
			i = 0;
			while (--i) {	/* delay >10ms for USB reset */
				nop();
				nop();
			}
		}
	}
}